#!/usr/bin/env python3
# coding:utf-8

# Copyright 2018 ISB Corp. <http://www.isb.co.jp/>
# All rights reserved.

import os
import isb_wsn_serial
import binascii
from intelhex import IntelHex
from time import sleep

VERSION = "1.0.5"

################################################################################
# defines
################################################################################

SECTOR_SIZE = 512        # 512 bytes
UDP_PORT_OTA = 31941

# UDP packet commands
OTACMD_REQ_OTAMODE_START   = 0x61
OTACMD_REQ_GET_VER         = 0x68
OTACMD_REQ_GET_BANKINFO    = 0x62
OTACMD_REQ_WRITE_START     = 0x40
OTACMD_REQ_WRITE_END       = 0x45
OTACMD_REQ_OTAMODE_STOP    = 0x64

OTACMD_RSP_OTAMODE_START   = 0x71
OTACMD_RSP_GET_VER         = 0x78
OTACMD_RSP_GET_BANKINFO    = 0x72
OTACMD_RSP_WRITE_START     = 0x70
OTACMD_RSP_WRITE_END       = 0x75
OTACMD_RSP_OTAMODE_STOP    = 0x74

# Response results
OTA_STATUS_PARAM_ERR      = 0x05
OTA_STATUS_ACK            = 0x06
OTA_STATUS_CHECKSUM_ERR   = 0x07
OTA_STATUS_NACK           = 0x15
OTA_STATUS_WRITE_ERR      = 0x1C
OTA_STATUS_NOWRITE        = 0x1D
OTA_STATUS_CHKSUM_ERR_ALL = 0x1E

################################################################################
# for debug
################################################################################
EnableDebugUdp = False

################################################################################

def calc_checksum_otacmd(data):
	checksum = 0
	for tmp in data:
		checksum += tmp

	checksum = ~checksum + 1;
	checksum &= 0xFF
	return checksum

def check_recv_cmd(udpData):
	com = 0
	comData = []

	if( len(udpData) == 0):
		print(" OTA Update Packet: Receive Timeout")
		return 0x00, com, comData
	elif(udpData[0] == 0x01) and (len(udpData) >= 6): # SOH
		dataLen = udpData[1]
		com = udpData[2]
		comData = udpData[3:-2]
		chksum = udpData[-2]
		#print("com:{:s} data:{:s} chk:{:s}".format(str(com), str(comData), str(chksum)))
		calcSum = calc_checksum_otacmd(udpData[1:-2])
		if(chksum != calcSum):
			print("check_recv_cmd(): detect CheckSum Error valid={:02X} != {:02X}".format(calcSum, chksum))

		strData = "".join(format(x, "02X")for x in comData[:])
		print(" OTA Update Packet: {:02X} {:02X} {:02X} {:s} {:02X} {:02X}".format(
			udpData[0],dataLen,com,strData,chksum,udpData[-1])
		)
		return 0x01, com, comData
	elif(udpData[0] == 0x02) and (len(udpData) >= 13): # STX
		com = int.from_bytes(udpData[1:3], byteorder='big') # DATANUM
		dataLen = int.from_bytes(udpData[3:5], byteorder='big') # LEN
		comData = udpData[5:-2]
		chksum = udpData[-2]
		calcSum = calc_checksum_otacmd(udpData[1:-2])
		if(chksum != calcSum):
			print("check_recv_cmd(): detect CheckSum Error valid={:02X} != {:02X}".format(calcSum, chksum))

		strData = "".join(format(x, "02X")for x in comData[:])
		print(" OTA Update Packet: {:02X} {:04X} {:04X} {:s} {:02X} {:02X}".format(
			udpData[0],com,dataLen,strData,chksum,udpData[-1])
		)
		return 0x02, com, comData
	else:
		comData = udpData[:]
		if(len(comData)>0):
			strRecvData = "".join(format(x, "02X")for x in comData)
			print(" OTA Update Packet: {:s}".format(strRecvData))

		return 0x00, com, comData

def build_ota_ctrlpkt(cmdCode, cmdInfo = None):
	if cmdInfo == None:
		cmdInfo = []

	ctrlCmd = [0x01] # SOH
	ctrlCmd += [(len(cmdInfo)+1)] # LEN
	ctrlCmd += [cmdCode] # COM
	checksum = calc_checksum_otacmd((ctrlCmd + cmdInfo)[1:]) # not include SOH
	ctrlCmd += cmdInfo
	ctrlCmd += [checksum]
	ctrlCmd += [0x03] # ETX
	return bytes(ctrlCmd)

def build_ota_datapkt(numOfSector, sectorData, footer):
	dataFrame = [0x02] # STX
	dataFrame += [(numOfSector>>8)&0xFF]
	dataFrame += [numOfSector&0xFF]
	dataFrame += [(len(sectorData)>>8)&0xFF]
	dataFrame += [len(sectorData)&0xFF]
	dataFrame += sectorData
	checksum = calc_checksum_otacmd(dataFrame[1:]) # not include STX
	dataFrame += [checksum]
	dataFrame += [footer] # ETX
	return bytes(dataFrame)

def build_datalist_fromhex(hexfile):
	ih = IntelHex()		# create empty object
	ih.loadhex(hexfile)	# load from hex
	segments = ih.segments()
	if segments == None:
		return [],0,0

	pAddr = 0
	prg_data_list = []
	for s in segments:
		startAddr = s[0]
		lastAddr = s[1]-1
		for addr in range(startAddr, lastAddr, SECTOR_SIZE):
			if(pAddr+SECTOR_SIZE >= ih.maxaddr()):
				break # already append
			else:
				sectorArray = ih.tobinarray(start=addr, size=SECTOR_SIZE)
				dataNum = ((addr - ih.minaddr()) / SECTOR_SIZE) + 1
				prg_data_list.append([int(dataNum), sectorArray])
				pAddr = addr

	return prg_data_list,ih.minaddr(),ih.maxaddr()

def input_hex_file(hexFile=None):
	input_str = hexFile
	file_path = None

	if(hexFile == None):
		print('Please input file path')
		print('\tex: bank0.hex')

	while(file_path==None):
		if(input_str != None):
			if(os.path.exists(input_str)):
				file_path = input_str
				break
			else:
				print("Path not found!! \'{0}\'\n".format(input_str))

		input_str = input('>>> ');

	return file_path

class WiSunOTA:
	def __init__(self,_wsn_ser,_mac):
		self.destIpv6 = isb_wsn_serial.mac2ipv6(_mac)
		self.wsnSerial = _wsn_ser
		self.fwData = []
		self.fwPath = ""
		self.minaddr = 0
		self.maxaddr = 0
		self.banknum = 0

	def __str__(self):
		str =  'class(WiSunOTA) v{0}\n'.format(VERSION)
		str += ' destination IPv6 : {:s}\n'.format(self.destIpv6)
		str += ' FW Data \'{:s}\''.format(self.fwPath)
		str += '  min:{:04X},max:{:04X},NumOfSecotr:{:d}' \
			.format(self.minaddr,self.maxaddr,len(self.fwData))
		return str

	def set_ip_from_mac(self, _mac):
		self.destIpv6 = isb_wsn_serial.mac2ipv6(_mac)
		print('set destination IPv6 : {:s}'.format(self.destIpv6))

	def load_fw_data(self, filePath):
		self.fwData,self.minaddr,self.maxaddr = build_datalist_fromhex(filePath)
		self.fwPath = filePath
		print("set FW Data \'{:s}\'".format(self.fwPath))
		print(" min:{:04X},max:{:04X},NumOfSecotr:{:d}" \
			.format(self.minaddr,self.maxaddr,len(self.fwData)))

	def send(self, cmdDoc, ctrlCmd):
		self.wsnSerial.flush_recv_buf()

		ret = self.wsnSerial.cmd_com_send_data(
			self.destIpv6,UDP_PORT_OTA,UDP_PORT_OTA,ctrlCmd )
		print(cmdDoc, end=":")
		if(ret == 0): print("Success"); return 0
		else: print("Failure"); return 1

	# for OTA Client
	def recv_for_client(self):
		cntNoData = 0
		while True:
			if( self.wsnSerial.isSleepingDevice ):
				self.wsnSerial.do_cmd_han_poll_req()

			recvData = self.wsnSerial.recv()
			if(len(recvData)==0):
				# timeout
				return 1

			cmdCode = isb_wsn_serial.get_cmd_code(recvData)
			cmdData = isb_wsn_serial.get_cmd_data(recvData)
			strCmdData = "".join(format(x, "02X")for x in cmdData)
			if(cmdCode == 0x6033):
				print("Notify OTA Operation Initiation: OTA Server={:s}".format(strCmdData))
			elif(cmdCode == 0x6034):
				print("Notify OTA Operation Termination: result={:s} OTA Server={:s}".format(
					strCmdData[:2],strCmdData[1*2:17*2]))
				if(cmdData[0] == 0x01): print(" Version upgrade succeeded")
				elif(cmdData[0] == 0x02): print(" Version upgrade failed")
				elif(cmdData[0] == 0x03): print(" No version upgrade")
				# OTA sequence complete.
				break
#			elif(cmdCode == 0x6037):
#				print("Notify HAN Indirect Queue Transmission: result={:s}, destIP={:s}".format(
#					strCmdData[:2],strCmdData[1*2:17*2]))
#			elif(cmdCode == 0x6036):
#				print("Notify HAN Indirect Queue Discard: num={:s}".format(strCmdData))
			elif(cmdCode == 0x2061):
				if(cmdData[0] == 0x01) and (cmdData[1] == 0x00):
					cntNoData += 1
				elif(cmdData[0] == 0x01) and (cmdData[1] == 0x01):
					cntNoData = 0

				if(cntNoData >= 1) and (cntNoData <= 5): sleep(0.2*cntNoData)
				elif(cntNoData > 5): sleep(2.0)
				else: sleep(0.2)

		self.wsnSerial.flush_recv_buf()
		return 2

	# for OTA Server
	def recv_for_server(self):
		while True:
			if( self.wsnSerial.isSleepingDevice ):
				self.wsnSerial.do_cmd_han_poll_req()

			recvData = self.wsnSerial.recv()
			if(len(recvData)==0): return recvData

			cmdCode = isb_wsn_serial.get_cmd_code(recvData)
			if(cmdCode == 0x6018):
				notifData = isb_wsn_serial.get_cmd_data(recvData)
				destPort = int.from_bytes(notifData[18:20], byteorder='big')
				udpDataLen = int.from_bytes(notifData[25:27], byteorder='big')
				if(destPort == UDP_PORT_OTA):
					break
				else:
					print(" UDP from port{:d}. size={:d}".format(destPort,udpDataLen))
					continue
			else: continue

		udpData = []
		if(cmdCode == 0x6018):
			notifData = isb_wsn_serial.get_cmd_data(recvData)
			srcIpv6 = notifData[0:16]
			strSrcIpv6 = "".join(format(x, "02X")for x in srcIpv6)
			srcPort = int.from_bytes(notifData[16:18], byteorder='big')
			destPort = int.from_bytes(notifData[18:20], byteorder='big')
			panId = int.from_bytes(notifData[20:22], byteorder='big')
			destAddrKind = notifData[22]
			security = notifData[23]
			rssi = notifData[24]
			udpDataLen = int.from_bytes(notifData[25:27], byteorder='big')
			udpData = notifData[27:]
			if(EnableDebugUdp):
				strUdpData = "".join(format(x, "02X")for x in udpData)
				print(" UDP from {:s}, PAN-ID={:04X}, RSSI={:d}dBm Data:{:s}" \
					.format(strSrcIpv6,panId,(rssi-0x100), strUdpData))

		# wait for the response of poll request
		if( self.wsnSerial.isSleepingDevice ):
			while True:
				recvData = self.wsnSerial.recv()
				if(len(recvData)==0): return recvData

				cmdCode = isb_wsn_serial.get_cmd_code(recvData)
				if(cmdCode == 0x2061):break

		return udpData

	# Start OTA Mode(cmd=0x61->0x71)
	def otamode_start(self):
		ctrlCmd = build_ota_ctrlpkt(OTACMD_REQ_OTAMODE_START)
		ret = self.send("Start OTA Mode Request [SEND]", ctrlCmd)
		if(ret != 0): return ret

		# response
		recvData = self.recv_for_server()
		header,com,comData = check_recv_cmd(recvData)
		if(header == 0x01) and (com == OTACMD_RSP_OTAMODE_START) \
 			and (comData[0] == OTA_STATUS_ACK):
			print("Start OTA Mode Response [RECV:Success]"); return 0

		print("Start OTA Mode Response [RECV:Failuer]"); return 1

	# Get OTA Version Information(cmd=0x68->0x78)
	def get_ver(self):
		ctrlCmd = build_ota_ctrlpkt(OTACMD_REQ_GET_VER)
		ret = self.send("Get OTA Version Information Request [SEND]", ctrlCmd)
		if(ret != 0): return ret

		# response
		recvData = self.recv_for_server()
		header,com,comData = check_recv_cmd(recvData)
		if(header == 0x01) and (com == OTACMD_RSP_GET_VER) \
 			and (comData[0] == OTA_STATUS_ACK):
			print("Get OTA Version Information Response [RECV:Success]")
			swId = int.from_bytes(comData[1:3], byteorder='big')
			swRev = int.from_bytes(comData[5:9], byteorder='big')
			print(" Firmware Version:{:04X}-{:02X}-{:02X}-{:08X}".format(
 				swId, comData[3], comData[4], swRev) )
			return 0, swRev

		print("Get OTA Version Information Response [RECV:Failuer]"); return 1, 0

	# Get OTA Write BANK Information(cmd=0x62->0x72)
	def get_bank_info(self):
		ctrlCmd = build_ota_ctrlpkt(OTACMD_REQ_GET_BANKINFO)
		ret = self.send("Get OTA Write BANK Information Request [SEND]", ctrlCmd)
		if(ret != 0): return ret

		# response
		recvData = self.recv_for_server()
		header,com,comData = check_recv_cmd(recvData)
		if(header == 0x01) and (com == OTACMD_RSP_GET_BANKINFO) \
 			and (comData[0] == OTA_STATUS_ACK):
			print("Get OTA Write BANK Information Response [RECV:Success]")
			print(" OTA Write BANK = BANK{:d}".format(comData[1]))
			self.banknum = comData[1]
			return 0
		else:
			print("Get OTA Write BANK Information Response [RECV:Failuer]"); return 1

	# Start OTA Write(cmd=0x40->0x70)
	def write_start(self,startAddr=None, endAddr=None):
		if( startAddr == None): startAddr = self.minaddr
		if( endAddr == None): endAddr = self.maxaddr

		cmdData = list(startAddr.to_bytes(4, 'big'))
		cmdData += list(endAddr.to_bytes(4, 'big'))
		ctrlCmd = build_ota_ctrlpkt(OTACMD_REQ_WRITE_START, cmdData)
		ret = self.send("Start OTA Write Request [SEND]", ctrlCmd)
		if(ret != 0): return ret

		# response
		recvData = self.recv_for_server()
		header,com,comData = check_recv_cmd(recvData)
		if(header == 0x01) and (com == OTACMD_RSP_WRITE_START) \
 			and (comData[0] == OTA_STATUS_ACK):
			print("Start OTA Write Response [RECV:Success]"); return 0
		else:
			print("Start OTA Write Rseponse [RECV:Failuer]"); return 1

	def send_all_otadata(self):
		for fwIndex, sector in enumerate(self.fwData):
			footer = 0x17 if(fwIndex < len(self.fwData)-1) else 0x03
			print("OTA Data Write {:.2f}%({:d}/{:d})".format(
				(1.0*(fwIndex+1)/len(self.fwData))*100,
				(fwIndex+1), len(self.fwData)))
			for i in range(3):
				ret = self.send_otadata(sector[0],sector[1],footer)
				if(ret==0): break
				else: self.wsnSerial.flush_recv_buf(); sleep(0.5)

			if(ret != 0): return 1

		return 0

	# OTA Data Write
	def send_otadata(self, numOfSector, sectorData, footer):
		dataFrame = build_ota_datapkt(numOfSector, sectorData, footer)
		ret = self.send("OTA Data Write Request [SEND(DataNum={:d})]".format(numOfSector), dataFrame)
		if(ret != 0): return ret

		# response
		recvData = self.recv_for_server()
		header,com,comData = check_recv_cmd(recvData)
		if(header == 0x02):
			dataNum = com
			status1 = comData[0]
			status2 = comData[1]
			crc32 = int.from_bytes(comData[2:6], byteorder='big')
			validCRC32 = binascii.crc32(sectorData)
			if(dataNum == numOfSector) and (status1==OTA_STATUS_ACK) \
				and ((status2 == OTA_STATUS_ACK and crc32 == validCRC32) \
					or ( status2 == OTA_STATUS_NOWRITE ) ):
				print("OTA Data Write Response [RECV(DataNum={:d}):Success]".format(dataNum))
				return 0
			else:
				print("OTA Data Write Response [RECV(DataNum={:d}):Failuer]".format(dataNum))
				print(" Request : DataNum={:d},CRC32={:08X})" \
					.format(numOfSector, validCRC32) )
				print(" Response: DataNum={:d},ST1={:02X},ST2={:02X},CRC32={:08X})" \
					.format(dataNum, status1, status2, crc32) )
				return 1
		else:
			print("OTA Data Write Response [RECV:Failuer]"); return 1

	# End OTA Write(cmd=0x45->0x75)
	def write_end(self):
		ctrlCmd = build_ota_ctrlpkt(OTACMD_REQ_WRITE_END)
		ret = self.send("End OTA Write Request [SEND]", ctrlCmd)
		if(ret != 0): return ret

		# response
		recvData = self.recv_for_server()
		header,com,comData = check_recv_cmd(recvData)
		if(header == 0x01) and (com == OTACMD_RSP_WRITE_END) \
			and (comData[0] == OTA_STATUS_ACK):
			print("End OTA Write Response [RECV:Success]"); return 0

		print("End OTA Write Response [RECV:Failuer]"); return 1

	# End OTA Mode(cmd=0x64->0x74)
	def otamode_end(self):
		ctrlCmd = build_ota_ctrlpkt(OTACMD_REQ_OTAMODE_STOP)
		ret = self.send("End OTA Mode Request [SEND]", ctrlCmd)
		if(ret != 0): return ret

		# response
		recvData = self.recv_for_server()
		header,com,comData = check_recv_cmd(recvData)
		if(header == 0x01) and (com == OTACMD_RSP_OTAMODE_STOP) \
 			and (comData[0] == OTA_STATUS_ACK):
			print("End OTA Mode Response [RECV:Success]"); return 0

		print("End OTA Mode Response [RECV:Failuer]"); return 1
