#!/usr/bin/env python3
# coding:utf-8

# Copyright 2018 ISB Corp. <http://www.isb.co.jp/>
# All rights reserved.

import sys
import serial
import binascii
import datetime
from serial.tools import list_ports
from time import sleep
import time

VERSION = "1.2.1"

################################################################################
# defines
################################################################################
UNIQUECODE_REQUEST  = [0xD0, 0xEA, 0x83, 0xFC]
UNIQUECODE_RESPONSE = [0xD0, 0xF9, 0xEE, 0x5D]
UNIQUECODE_NOTIFY   = [0xD0, 0xF9, 0xEE, 0x5D]

# Operation mode
HAN_MODE_PANC = 0x01
HAN_MODE_CRDI = 0x02
HAN_MODE_DEV  = 0x03
HAN_MODE_DUAL = 0x05
# Transmission power
TX_POWER_20MW = 0x00
TX_POWER_10MW = 0x01
TX_POWER_1MW  = 0x02

COMMAND_SUCCESS = 0x01

# Result of data transmission
UDP_SEND_SUCCESS          = 0x00
UDP_SEND_FAIL_360S_LIMIT  = 0x02
UDP_SEND_FAIL_CCA         = 0x03
UDP_SEND_FAIL_NACK        = 0x05
UDP_SEND_FAIL_OTHER       = 0x08
UDP_SEND_QUEUING          = 0x0F
UDP_SEND_NOT_QUEUING      = 0x00
UDP_SEND_QUEUING_SUCCESS  = 0x10
UDP_SEND_QUEUING_FAIL     = 0x20

UDP_SEND_ID_SUCCESS       = 0x1F
UDP_SEND_ID_MULTI_SUCCESS = 0x10

################################################################################
# for debug
################################################################################
EnableDebugSerial = True

################################################################################

g_recv_buf = []
def recv_cmd_advanced(serial_obj):
	cmd_buf = []
	gb_buf = []
	global g_recv_buf
	error_flg = False

	while(True):
		if( len(g_recv_buf) < 4):
			read_buf = serial_obj.read(1)
			if( len(read_buf) == 0):
				# timeout
				error_flg = True
				break
			else:
				g_recv_buf += read_buf
				continue

		# find unique code
		idxUni = 0
		while( idxUni < len(UNIQUECODE_RESPONSE) ):
			if( len(g_recv_buf) < len(UNIQUECODE_RESPONSE) ):
				break

			if( g_recv_buf[idxUni] != UNIQUECODE_RESPONSE[idxUni]):
				gb_buf += g_recv_buf[:idxUni+1]
				g_recv_buf = g_recv_buf[idxUni+1:]
				idxUni = 0
			else:
				idxUni += 1

		if( idxUni != len(UNIQUECODE_RESPONSE) ):
			continue

		if( len(g_recv_buf) < 12):
			g_recv_buf += serial_obj.read( 12 - len(g_recv_buf) )

		break # end of loop

	if( len(g_recv_buf) < 12 ) or ( error_flg ):
		#print("recv_cmd:[err] timeout. header(len={:d})".format(len(g_recv_buf)))
		gb_buf += g_recv_buf[:]
		cmd_buf = []
		error_flg = True
	else:
		# header reception was complete.
		cmd_buf = g_recv_buf[:12+1]

	if( not error_flg ):
		head_chksum = 0
		for byte in g_recv_buf[:8]:
			head_chksum += byte

		recv_chksum = int.from_bytes(g_recv_buf[8:10], byteorder='big')
		if(head_chksum != recv_chksum):
			print("recv_cmd:[err] header checksum err. {:04X}!={:04X}." \
				.format(head_chksum, recv_chksum))
			g_recv_buf = g_recv_buf[12:]
			error_flg = True

	if( not error_flg ):
		body_len = int.from_bytes(cmd_buf[6:8], byteorder='big') - 4
		total_cmd_len = 12 + body_len
		if( len(g_recv_buf) >= total_cmd_len ):
			cmd_buf = g_recv_buf[:total_cmd_len+1]
		else:
			req_len = total_cmd_len - len(cmd_buf)

			while(True):
				follow_data = serial_obj.read( req_len )
				g_recv_buf += follow_data

				if( len(follow_data) == 0) or (len(follow_data) == req_len):
					break
				else:
					req_len -= len(follow_data)
					continue
			cmd_buf = g_recv_buf[:]

		if(len(cmd_buf) < total_cmd_len):
			print("recv_cmd:[err] timeout. body(buf={:d},total={:d})" \
				.format(len(cmd_buf),total_cmd_len))
			g_recv_buf = g_recv_buf[12:]
			error_flg = True

	if( not error_flg ):
		body_chksum = 0
		for byte in cmd_buf[12:]:
			body_chksum += byte
		body_chksum &= 0xFFFF

		recv_chksum = int.from_bytes(cmd_buf[10:12], byteorder='big')
		if body_chksum != recv_chksum:
			print("recv_cmd:[err] body checksum err. {:04X}!={:04X}" \
				.format(body_chksum, recv_chksum))
			g_recv_buf = g_recv_buf[12:]
			#error_flg = True
		else:
			# body reception was complete.
			g_recv_buf = g_recv_buf[total_cmd_len:]

	if( error_flg ):
		g_recv_buf = []

	if( len(gb_buf) > 0 ):
		print("recv_cmd:[info] prefix garbage={:s}".format(
			"".join(format(x, "02X")for x in gb_buf)) )

	return bytes(cmd_buf)

def send_cmd(serial_obj, data):
	ret = serial_obj.write(data)
	return 0 if(len(data) == ret) else 1

def build_cmd(uniqueCode, cmdCode, cmdData):
	command = uniqueCode[:]
	body = []
	headerCheckSum = 0
	bodyCheckSum = 0

	# Command Code
	command += cmdCode.to_bytes(2, byteorder='big')

	# Length of Data
	cmdLen = len(cmdData) + 4
	command += cmdLen.to_bytes(2, byteorder='big')

	# Header Checksum
	for item in command:
		headerCheckSum += item
	headerCheckSum &= 0xFFFF
	command += headerCheckSum.to_bytes(2, byteorder='big')

	# Data Checksum
	cmdLen = len(cmdData)
	if(cmdLen > 0):
		body = cmdData[:]
		for item in body:
			bodyCheckSum += item
	bodyCheckSum &= 0xFFFF
	command += bodyCheckSum.to_bytes(2, byteorder='big')

	command += body
	return bytes(command)

def build_req_cmd(cmdCode, cmdData = None):
	if cmdData == None:
		cmdData = []
	return build_cmd(UNIQUECODE_REQUEST, cmdCode, cmdData)

def build_rsp_cmd(cmdCode, cmdData = None):
	if cmdData == None:
		cmdData = []
	return build_cmd(UNIQUECODE_RESPONSE, cmdCode, cmdData)

def get_cmd_code(cmd_buf):
	return int.from_bytes(cmd_buf[4:6], byteorder='big')

def get_cmd_data(cmd_buf):
	return cmd_buf[12:]

def input_dest_mac(addr=None):
	ret_addr = None
	input_str = addr

	if(addr==None):
		print('Please input destination MAC address(8byte).')
		print('\tex: 0123456789ABCDEF')

	while(ret_addr==None):
		if(input_str!=None)and(len(input_str) == 16):
			ret_addr = bytes.fromhex(input_str)
		else:
			input_str = input('>>> ');

	return ret_addr

# EUI64 -> IPv6
def mac2ipv6(eui64):
	ifId = bytearray(eui64[:])
	ifId[0] = ifId[0] ^ 0x02 # Universal/Local bit inversion
	strIfid = "".join(format(x, "02X")for x in ifId)
	ipv6 = "FE80000000000000%s" % (strIfid)
	return ipv6

def input_pana_pass(pana_pass=None):
	ret_pass = None
	input_str = pana_pass

	if(pana_pass==None):
		print('Please input PANA password(16byte).')
		print('\tex(ascii): 30313233343536373839414243444546')
		print('\tex(char): 0123456789ABCDEF')

	while(ret_pass == None):
		if(input_str!=None)and(len(input_str) == 32):
			ret_pass = bytes.fromhex(input_str)
		elif(input_str!=None)and(len(input_str) == 16):
			ret_pass = bytes([ord (s) for s in input_str])
		else:
			input_str = input('>>> ');

	return ret_pass

# serial communication class for Wi-SUN Dual Stack
class WiSunSerial():
	UART_BAUDRATE = 115200   # bps
	UART_TIMEOUT = 10.0      # second

	def __init__(self,_tty_dev, _rtscts=False):
		self.serialObj = serial.Serial()
		self.serialObj.port = _tty_dev
		self.serialObj.baudrate = self.UART_BAUDRATE
		self.serialObj.timeout = self.UART_TIMEOUT
		self.serialObj.rtscts = _rtscts
		try:
			self.serialObj.open()
		except:
			print("Port open error:{:s}".format(self.serialObj.port))
			print(" port list:")
			ports = list_ports.comports()
			for info in ports: print("\t",info)

		self.hanMode = HAN_MODE_PANC
		self.isSleepingDevice = False

	def __del__(self):
		self.serialObj.close()

	def isOpen(self):
		return self.serialObj.isOpen()

	def setTimeout(self, time):
		self.serialObj.timeout = time

	def __str__(self):
		str =  'class(WiSunSerial) v{0}\n'.format(VERSION)
		str += ' tty device  :{:s}\n'.format(self.serialObj.port)
		str += ' baudrate    :{:d}\n'.format(self.serialObj.baudrate)
		str += ' timeout     :{:f}\n'.format(self.serialObj.timeout)
		str += ' rtscts      :{0}\n'.format(self.serialObj.rtscts)
		return str

	def recv(self):
		recvData = recv_cmd_advanced(self.serialObj)
		now = datetime.datetime.now()
		if(len(recvData) == 0):
			return recvData

		if(EnableDebugSerial):
			strRecv = "".join(format(x, "02X")for x in recvData)
			strNow = now.strftime("%Y/%m/%d %H:%M:%S.") + "%03d" % (now.microsecond // 1000)
			print(" {:s} Rx: {:s} {:s} {:s} {:s} {:s} {:s}".format(
				strNow,strRecv[0:8],strRecv[8:12],strRecv[12:16],
				strRecv[16:20],strRecv[20:24],strRecv[24:]))

		return recvData

	def flush_recv_buf(self):
		self.setTimeout(0.1)
		while(True):
			recvData = self.recv()
			if(len(recvData)==0):
				self.setTimeout(self.UART_TIMEOUT)
				return 0

	def send(self,data):
		ret = send_cmd(self.serialObj,data)
		now = datetime.datetime.now()

		if(EnableDebugSerial):
			strSend = "".join(format(x, "02X")for x in data)
			strNow = now.strftime("%Y/%m/%d %H:%M:%S.") + "%03d" % (now.microsecond // 1000)
			print(" {:s} Tx: {:s} {:s} {:s} {:s} {:s} {:s}".format(
				strNow,strSend[0:8],strSend[8:12],strSend[12:16],
				strSend[16:20],strSend[20:24],strSend[24:]))

		return ret

	def chk_response(self, cmdDoc, rspCmd, rspRet=None):
		recvData = self.recv()
		cmpCmd = build_rsp_cmd(rspCmd, [rspRet])
		print(cmdDoc, end=":")
		if(recvData == cmpCmd):
			print("Success"); return 0
		else:
			print("Failure"); return 1

	# Common commands:Get MAC Address(cmd=0x000E->0x200E)
	def cmd_com_get_addr(self):
		sendData = build_req_cmd(0x000E)
		self.send(sendData)
		# response
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		if(get_cmd_code(recvData) == 0x200E):
			if(cmdData[0] == COMMAND_SUCCESS):
				print("Get MAC Address: Success")
				strAddr = "".join(format(x, "02X")for x in cmdData[1:])
				print(" MAC address: {:s}".format(
	 				strAddr) )
				return 0

		print("Get MAC Address: Failure"); return 1

	# Common commands:Get Version Information(cmd=0x006B->0x206B)
	def cmd_com_get_ver(self):
		sendData = build_req_cmd(0x006B)
		self.send(sendData)
		# response
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		if(get_cmd_code(recvData) == 0x206B):
			if(cmdData[0] == COMMAND_SUCCESS):
				print("Get Version Information: Success")
				fwid = int.from_bytes(cmdData[1:3], byteorder='big')
				rev = int.from_bytes(cmdData[5:9], byteorder='big')
				print(" Firmware Version:{:04X}-{:02X}-{:02X}-{:08X}({:d})".format(
	 				fwid, cmdData[3], cmdData[4], rev, rev ) )
				return 0

		print("Get Version Information: Failure"); return 1

	# Common commands:Reset Hardware(cmd=0x00D9->0x6019)
	def cmd_com_reset(self):
		sendData = build_req_cmd(0x00D9)
		self.send(sendData)
		# response
		while(True):
			recvData = self.recv()
			cmdCode = get_cmd_code(recvData)
			if(cmdCode == 0x6019):
				break
			elif(cmdCode == 0x60DA):
				self.send(sendData)
			else:
				continue

		print("Reset Hardware: Success")
		return 0

	# Common commands:Set Initial Settings(cmd=0x005F->0x205F)
	def cmd_com_init(self,cmd_data):
		sendData = build_req_cmd(0x005F, cmd_data)
		self.send(sendData)
		# response
		ret = self.chk_response("Set Initial Settings",0x205F, COMMAND_SUCCESS)
		if(ret == 0):
			self.hanMode = cmd_data[0]
			if(cmd_data[1] == 0x01) and (self.hanMode == HAN_MODE_DEV):
				self.isSleepingDevice = True

		return ret

	# HAN commands:Initiate HAN Operation(cmd=0x000A->0x200A)
	def cmd_han_start(self,panid,panc_mac=None):
		if(self.hanMode == HAN_MODE_PANC) or (self.hanMode == HAN_MODE_DUAL):
			cmd_data = panid.to_bytes(2, byteorder='big')
		elif( panc_mac == None ):	# Coordinator or End device
			cmd_data = [0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF]
		else:
			cmd_data = panc_mac

		sendData = build_req_cmd(0x000A, cmd_data)
		self.send(sendData)

		# response
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		print("Initiate HAN Operation", end=":")
		if(get_cmd_code(recvData) == 0x200A):
			result = cmdData[0]
			if(result == COMMAND_SUCCESS):
				channel = cmdData[1]
				panid = int.from_bytes(cmdData[2:4], byteorder='big')
				print("Success")
				strMac = "".join(format(x, "02X")for x in cmdData[4:12])
				if(len(strMac) > 0):
					rssi = cmdData[12] - 256
					print(" MAC Connected to \'{:s}\', rssi={:d}dBm".format(strMac,rssi))

				return 0

		print("Failure"); return 1

	# HAN commands:Switch HAN Acceptance Connection Mode(cmd=0x0025->0x2025)
	def cmd_han_switch_connect_mode(self,mode):
		sendData = build_req_cmd(0x0025, [mode])
		self.send(sendData)
		# response
		ret = self.chk_response("Switch HAN Acceptance Connection Mode",0x2025,
 				COMMAND_SUCCESS)
		return ret

	# HAN commands:Set HAN PANA Authentication Information(cmd=0x002C->0x202C)
	def cmd_han_pana_setting(self, mac, password):
		if(self.hanMode == HAN_MODE_PANC) or (self.hanMode == HAN_MODE_DUAL):
			cmd_data = mac[:]
			cmd_data += password[:]
		else:	# Coordinator or End device
			cmd_data = password

		sendData = build_req_cmd(0x002C, cmd_data)
		self.send(sendData)
		# response
		ret = self.chk_response("Set HAN PANA Authentication Information",0x202C,
 				COMMAND_SUCCESS)
		return ret

	# HAN commands:Get HAN Group Key(cmd=0x0028->0x2028)
	def cmd_han_get_group_key(self):
		sendData = build_req_cmd(0x0028)
		self.send(sendData)
		# response
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		if(get_cmd_code(recvData) == 0x2028):
			if(cmdData[0] == COMMAND_SUCCESS):
				print("Get HAN Group Key: Success")
				groupKey = "".join(format(x, "02X")for x in cmdData[1:17])
				keyId = cmdData[17]
				print(" HAN Key: {:02X}, {:s}".format(keyId,groupKey) )
				return 0

		print("Get HAN Group Key: Failure")
		return 1

	# HAN commands:Initiate HAN PANA(cmd=0x003A->0x203A)
	def cmd_han_pana_start(self):
		sendData = build_req_cmd(0x003A)
		self.send(sendData)
		# response
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		print("Initiate HAN PANA", end=":")
		if(get_cmd_code(recvData) == 0x203A):
			if(cmdData[0] == COMMAND_SUCCESS):
				print("Success.");
				if(self.hanMode != HAN_MODE_PANC) and (self.hanMode != HAN_MODE_DUAL):
					# Coo(dinator or End device
					strPancMac = "".join(format(x, "02X")for x in cmdData[1:])
					print(" PANA auth started with \'{:s}\'".format(strPancMac))

				return 0

		print("Failure")
		return 1

	# OTA update commands:Initiate OTA Client(cmd=0x0201->0x2201)
	def cmd_ota_client_start(self):
		sendData = build_req_cmd(0x0201)
		self.send(sendData)
		# response
		ret = self.chk_response("Initiate OTA Client",0x2201,COMMAND_SUCCESS)
		return ret

	# OTA update commands:Terminate OTA Client(cmd=0x0202->0x2202)@TODO
	def cmd_ota_client_end(self):
		sendData = build_req_cmd(0x0202)
		self.send(sendData)
		# response
		ret = self.chk_response("Terminate OTA Client",0x2202,
 				COMMAND_SUCCESS)
		return ret

	# Common commands:Open UDP Port(cmd=0x0005->0x2005)
	def cmd_com_open_port(self,udp_port):
		cmd_data = udp_port.to_bytes(2, byteorder='big')
		sendData = build_req_cmd(0x0005, cmd_data)
		self.send(sendData)
		# response
		ret = self.chk_response("Open UDP Port({:d})".format(udp_port),
				0x2005,COMMAND_SUCCESS)
		return ret

	# Common commands:Close UDP Port(cmd=0x0006->0x2006)
	def cmd_com_close_port(self,udp_port):
		cmd_data = udp_port.to_bytes(2, byteorder='big')
		sendData = build_req_cmd(0x0006, cmd_data)
		self.send(sendData)
		# response
		ret = self.chk_response("Close UDP Port({:d})".format(udp_port),
				0x2006,COMMAND_SUCCESS)
		return ret

	# Common commands:Transmit Data(cmd=0x0008->0x2008->0x6037 or 0x6038)
	def cmd_com_send_data(self,ipv6,src_port,dest_port,dataArray):
		dataArrayLen = len(dataArray)
		reqSendData = "%s%04X%04X%04X" % (
			ipv6,
 			src_port,
 			dest_port,
			dataArrayLen
		)
		cmd_data = bytes.fromhex(reqSendData)
		cmd_data += dataArray
		sendData = build_req_cmd(0x0008, cmd_data)
		self.send(sendData)

		# response
		preTime = time.perf_counter()
		while(True):
			recvData = self.recv()
			cmdCode = get_cmd_code(recvData)
			cmdData = get_cmd_data(recvData)
			if(cmdCode == 0x2008):
				break

			deltaSec = time.perf_counter() - preTime
			serialTO = self.serialObj.timeout
			if(deltaSec > serialTO):
				print("Transmit Data:Failure, Response Timeout({:.1f}sec)".format(serialTO))
				return 3

		flg_failure = True
		cmdResp = cmdData[0]
		if(cmdCode == 0x2008) and (len(cmdData) >= 2):
			sendRslt = cmdData[1]
			part = cmdData[2:]
			if(cmdResp == COMMAND_SUCCESS):
				if( sendRslt == UDP_SEND_SUCCESS ):
					flg_failure = False
				elif(sendRslt == UDP_SEND_ID_SUCCESS ):
					flg_failure = False
				elif(sendRslt == UDP_SEND_ID_MULTI_SUCCESS ):
					flg_failure = False

		if( flg_failure ):
			print("Transmit Data:Failure", end=", ")
			if(len(cmdData) >= 2):
				print("cmdResp={:02X}, SendRslt={:02X}".format(cmdResp, sendRslt))
				self.print_result_detail_tx(sendRslt)
				self.print_result_detail_queuing(sendRslt)

				if( (sendRslt&0x0F) == UDP_SEND_FAIL_360S_LIMIT):
					return 2
			else:
				print("cmdResp={:02X}".format(cmdResp))

			return 1

		if(sendRslt != UDP_SEND_ID_SUCCESS):
			#print("Transmit Data:Success")
			return 0

		print("\tWaiting for send indirect result notification...")
		numOfFragment = 1
		while( numOfFragment >= 1 ):
			recvData = self.recv()
			cmdCode = get_cmd_code(recvData)
			cmdData = get_cmd_data(recvData)
			if(cmdCode == 0x6037):
				numOfFragment = cmdData[1]
				if( cmdData[0] == 0x00 ): break
				else:
					destIp = cmdData[2:18]
					strDestIp = "".join(format(x, "02X")for x in destIp)
					print("Transmit Data(Indirect):Failure(result={:02X}, dest={:s})" \
						.format(cmdData[0],strDestIp))
					# print of failuer factors
					self.print_result_detail_tx(cmdData[0])
			elif(cmdCode == 0x6036):
				print("Transmit Data(Indirect):Failure. del {:d} packets".format(cmdData[0]))
				return 1
			else:
				continue

		#print("Transmit Data(Indirect):Success")
		return 0

	def print_result_detail_tx( self, resltTx ):
		detail = resltTx&0x0F
		print("\tDetailed result of UDP data transmission (Z)")
		if(detail == UDP_SEND_SUCCESS):
			print("\t\t0x{:02X}:Succeeded".format(detail))
		elif(detail == UDP_SEND_FAIL_360S_LIMIT):
			print("\t\t0x{:02X}:Transmission failed due to limitation on the sum of transmission data amount".format(detail))
		elif(detail == UDP_SEND_FAIL_CCA):
			print("\t\t0x{:02X}:Transmission failed due to failure in CCA".format(detail))
		elif(detail == UDP_SEND_FAIL_NACK):
			print("\t\t0x{:02X}:Transmission failed due to unreceived ACK".format(detail))
		elif(detail == UDP_SEND_FAIL_OTHER):
			print("\t\t0x{:02X}:Transmission failed due to other cause of failure".format(detail))
		elif(detail == UDP_SEND_QUEUING):
			print("\t\t0x{:02X}:No transmission (only queuing)".format(detail))
		else:
			print("\t\t0x{:02X}:unkown result".format(detail))

	def print_result_detail_queuing( self, resltTx ):
		detail = resltTx&0xF0
		print("\tDetailed result of indirect queuing(Y)")
		if(detail == UDP_SEND_NOT_QUEUING):
			print("\t\t0x{:02X}:No data queued in indirect queue".format(detail))
		elif(detail == UDP_SEND_QUEUING_SUCCESS):
			print("\t\t0x{:02X}:Data queued in indirect queue".format(detail))
		elif(detail == UDP_SEND_QUEUING_FAIL):
			print("\t\t0x{:02X}:Failed to queue data in indirect queue".format(detail))
		else:
			print("\t\t0x{:02X}:unkown result".format(detail))


	# for PAN Coordinator
	def wait_for_connect(self,clientAddr):
		while(True):
			recvData = self.recv()
			if(len(recvData) == 0):
				continue

			cmdCode = get_cmd_code(recvData)
			cmdData = get_cmd_data(recvData)
			if(cmdCode == 0x601A): # Notify Connection Status Change(0x601A)
				state = cmdData[0]
				mac = cmdData[1:9]
				strMac = "".join(format(x, "02X")for x in mac)
				rssi = cmdData[9] - 256
				if (state == 0x01): # MAC connection completed
					print("Notify MAC connection \'{:s}\'".format(strMac))
				elif (state == 0x02): # PANA connection completed
					print("Notify PANA auth completion \'{:s}\'".format(strMac))
					if( mac == clientAddr): return 0
				elif (state == 0x03): # MAC disconnected
					print("Notify MAC disconnection \'{:s}\'".format(strMac))
				elif (state == 0x04): # PANA disconnected
					print("Notify PANA disconnection \'{:s}\'".format(strMac))
				continue
			elif(cmdCode == 0x6023): # Notify HAN Acceptance Connection Mode Change(0x6023)
				print("The module has returned to normal mode.\nPlease restart this script.")
				return 1
			else:
				continue

		return 0

	# for Coodinator or end device
	def wait_for_pana_auth(self):
		while(True):
			if( self.isSleepingDevice ):
				self.do_cmd_han_poll_req()

			recvData = self.recv()
			if(len(recvData) == 0):
				continue

			cmdCode = get_cmd_code(recvData)
			cmdData = get_cmd_data(recvData)
			if(cmdCode == 0x6028): # Notify PANA Authentication Result(0x6028)
				panaRslt = cmdData[0]
				mac = cmdData[1:9]
				strMac = "".join(format(x, "02X")for x in mac)
				if (panaRslt == 0x01): #  Authentication succeeded
					print("PANA Auth Complete \'{:s}\'".format(strMac))
					break
				else:
					print("PANA Auth failed(result={:02X}).".format(panaRslt))
					return 1
			elif(cmdCode == 0x2061):
				sleep(0.2)
				continue
			else:
				continue

		if( self.isSleepingDevice ):
			while(True):
				recvData = self.recv()
				if(get_cmd_code(recvData) == 0x2061):
					break

		return 0

	# HAN commands:Transmit HAN Poll Request(cmd=0x0061->0x2061)
	def cmd_han_poll_req(self):
		self.do_cmd_han_poll_req()
		return self.do_cmd_han_poll_res()

	def do_cmd_han_poll_req(self):
		sendData = build_req_cmd(0x0061)
		return self.send(sendData)

	def do_cmd_han_poll_res(self):
		# response
		while(True):
			recvData = self.recv()
			if(len(recvData) == 0):
				self.do_cmd_han_poll_req()
				continue
			if(get_cmd_code(recvData) == 0x2061):
				break

		cmdData = get_cmd_data(recvData)
		if(cmdData[0] == COMMAND_SUCCESS):
			# print("Transmit HAN Poll Request: Success")
			return 0, cmdData
		else:
			print("Transmit HAN Poll Request: Failure")
			return 1, cmdData

